#!/bin/python
"""
@copyright: IBM
"""

HEADERS = {
            "Content-Type":"application/json",
            "Accept":"application/json"
        }

API_HEADERS = {
            "Content-Type":"application/json",
            "Accept":"application/json"
        }

CONFIG_YAML_ENV_VAR = "IVIA_CONFIG_YAML"
LEGACY_CONFIG_YAML_ENV_VAR = "ISVA_CONFIG_YAML"


CONFIG_YAML = "config.yaml"

CONFIG_BASE_DIR = "IVIA_CONFIG_BASE"
LEGACY_CONFIG_BASE_DIR = "ISVA_CONFIG_BASE"

KUBERNETES_CONFIG = "IVIA_KUBERNETES_YAML_CONFIG"
LEGACY_KUBERNETES_CONFIG = "ISVA_KUBERNETES_YAML_CONFIG"

KUBERNETES_CLIENT_SLEEP = "IVIA_KUBERNETES_RESTART_SLEEP"
LEGACY_KUBERNETES_CLIENT_SLEEP = "ISVA_KUBERNETES_RESTART_SLEEP"

DOCKER_COMPOSE_CONFIG = "IVIA_DOCKER_COMPOSE_CONFIG"
LEGACY_DOCKER_COMPOSE_CONFIG = "ISVA_DOCKER_COMPOSE_CONFIG"

MGMT_USER_ENV_VAR = "IVIA_MGMT_USER"
LEGACY_MGMT_USER_ENV_VAR = "ISVA_MGMT_USER"

MGMT_PWD_ENV_VAR = "IVIA_MGMT_PWD"
LEGACY_MGMT_PWD_ENV_VAR = "ISVA_MGMT_PWD"

MGMT_URL_ENV_VAR = "IVIA_MGMT_BASE_URL"
LEGACY_MGMT_URL_ENV_VAR = "ISVA_MGMT_BASE_URL"

MGMT_OLD_PASSWORD_ENV_VAR = "IVIA_MGMT_OLD_PWD"
LEGACY_MGMT_OLD_PASSWORD_ENV_VAR = "ISVA_MGMT_OLD_PWD"

LOG_LEVEL = "ISVA_CONFIGURATOR_LOG_LEVEL"
LEGACY_LOG_LEVEL = "IVIA_CONFIGURATOR_LOG_LEVEL"

MGMT_EXT_USER_ENV_VAR = "IVIA_EXT_USER"
MGMT_EXT_PWD_ENV_VAR = "IVIA_EXT_PWD"
